import { defineQuery, hasComponent } from 'bitecs'
import { query as q } from './query-parser/index.js'
import { parseAll } from './parser.js'
import { curry, of, is, reduce, flip, concat, map } from './fn.js'
import { Relationship } from './query-parser/types.js'

/*
 * necessary engine structure
 * worlds: World[]
 * Component: Component[]
 */

const prepareQuery = (query, component) => {
  const { match, returnValues } = query
  const relationship = is(Relationship, match)

  if (relationship) {
    const { from: f, to: t, edge: e } = match
    const [from, to, edge] = [f, t, e].map(q => prepareQuery({ match: q, returnValues }, component))

    return { from, to, edge, relationship, schema: query }
  } else {
    const label = match.label?.value?.value
    const name = match.name?.value?.value
    const type = component[label]

    return { name, label, type, relationship, query: defineQuery([type]), schema: query }
  }
}

const getEndNode = end => end.relationship ? end.from : end

const matches = (world, type, query, set = new Set()) => entity => {
  if (set.has(entity)) {
    return true
  } else if (hasComponent(world, type, entity) && query.includes(entity)) {
    set.add(entity)
    return true
  } else {
    return false
  }
}

const queryRelationship = (from, edge, to, world) => {
  const matchesFrom = matches(world, from.type, from.results)
  const matchesTo = matches(world, to.type, to.results)

  const Edge = edge.type

  const result = []
  for (let i = 0; i < edge.results.length; i++) {
    const eid = edge.results[i]

    if (matchesFrom(Edge.from[eid]) && matchesTo(Edge.to[eid])) {
      result.push(eid)
    }
  }

  return result
}

const cat = flip(concat)

const executeQuery = (query, world) => {
  if (!query.relationship) {
    return map(id => ({ [id]: query }), query.query(world))
  }

  const { from, to: _to, edge } = query
  const to = getEndNode(_to)

  const Edge = edge.type

  const edges = queryRelationship(
    { type: from.type, results: from.query(world) },
    { type: Edge, results: edge.query(world) },
    { type: to.type, results: to.query(world) },
    world
  )

  if (_to.relationship) {
    return reduce((acc, eid) => {
      const next = {
        ..._to,
        from: {
          ...to,
          query: () => of(Edge.to[eid])
        }
      }

      const results = map(cat([eid]), executeQuery(next, world))
      const expanded = map(
        ([l, r]) => ({ [l]: query, ...r }),
        results
      )
      return [...acc, ...expanded]
    }, [], edges)
  } else {
    return map(id => of({ [id]: query }), edges)
  }
}

const val = v => v?.value?.value
const include = returns => name => name != null && returns.includes(name)

const resolveReturns = results => {
  return map(result => {
    return reduce((acc, [id, query]) => {
      const q = query.schema.match
      const inc = include(query.schema.returnValues.values.map(x => x.value))
      if (is(Relationship, q)) {
        const Edge = query.edge.type
        const fn = val(q.from.name)
        const en = val(q.edge.name)
        const tn = val(q.to.name)
        return {
          ...acc, ...{
            ...(inc(fn) && { [fn]: Edge.from[id] }),
            ...(inc(en) && { [en]: parseInt(id, 10) }),
            ...(inc(tn) && { [tn]: Edge.to[id] })
          }
        }
      } else {
        const name = val(q.name)
        return { ...acc, ...(inc(name) && { [name]: id }) }
      }
    }, {}, Object.entries(result))
  }, results)
}

export const query = curry((input, engine) => {
  return parseAll(q, input).map(({ use, ...rest }) => {

    const worldName = use && use.value || 'default'
    const world = engine.world[worldName]
    const prepared = prepareQuery(rest, engine.component)
    const edges = executeQuery(prepared, world)
    return resolveReturns(edges)
  })
})

